import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Random;

import com.lowagie.text.BadElementException;
import com.lowagie.text.Document;
import com.lowagie.text.DocumentException;
import com.lowagie.text.pdf.PdfWriter;


public class BLFGroupingGA2D {
	
	/** Maksimalna površina ostatka za koji se materijal smatra dovoljno dobro izrezanim */
	private int maxOstatak;
	/** Popis količina predmeta koji je potrebno izrezati */
	private HashMap<Integer, Integer> items;
	/** Popis predmeta koji je potrebno izrezati */
	private HashMap<Integer, Predmet> predmeti;
	/** Širina materijala */
	private int stockWidth;
	/** Visina materijala */
	private int stockHeight;
	/** Parametri genetskog algoritma */
	private ParametriGA parametri;
	/** Trenutna populacija rješenja genetskog algoritma */
	private Populacija populacija;
	/** Ukupni broj predmeta koji je potrebno izrezati */
	private int numberOfItems;
	/** Generator nasumičnih brojeva */
	private Random random;
	
	/**
	 * Konstruktor objekta
	 * 
	 * @param ulaz datoteka sa ulaznim podacima
	 * @param parametri datoteka sa parametrima algoritma
	 * @param time trenutno vrijeme, dobiveno iz Date objekta
	 */
	public BLFGroupingGA2D(File ulaz, File parametri, long time){
		super();
		try {
			this.items = new HashMap<Integer, Integer>();
			this.predmeti = new HashMap<Integer,Predmet>();
			this.random = new Random(time);
			this.numberOfItems = 0;
			unos(ulaz);
			this.parametri = new ParametriGA(parametri);
		} catch (IOException e) {
			e.printStackTrace();
		}
		this.populacija = new Populacija(generiranjePopulacije());
	}
	
	/**
	 * Učitava podatke iz ulazne datoteke
	 * 
	 * @param ulaz ulazna datoteka iz koje se učitavaju ulazni podaci
	 * @throws IOException greška prilikom čitanja iz datoteke
	 */
	private void unos(File ulaz) throws IOException{
		BufferedReader citac = new BufferedReader(new FileReader(ulaz));
		String red = citac.readLine();
		String[] temp = red.split("\t");
		stockWidth = Integer.valueOf(temp[0]);
		stockHeight = Integer.valueOf(temp[1]);
		maxOstatak = Integer.valueOf(temp[2]);
		red = citac.readLine();
		int i = 1;
		while (red != null) {
			temp = red.split("\t");
			items.put(i, Integer.valueOf(temp[2]));
			predmeti.put(i, new Predmet(i, Integer.valueOf(temp[0]), Integer.valueOf(temp[1])));
			numberOfItems += Integer.valueOf(temp[2]);
			i++;
			red = citac.readLine();
		}
	}
	
	/**
	 * Funkcija koja postupkom slučajnog odabira generira jedinke, te tako stvara početnu populaciju
	 * 
	 * @return generirana populacija
	 */
	private ArrayList<Kromosom> generiranjePopulacije(){
		ArrayList<Kromosom> rjesenja = new ArrayList<Kromosom>();
		for (int i = 0 ; i < parametri.getVelicinaPopulacije(); i++ ){
			Kromosom jedinka = new Kromosom();
			int itemsLeft = numberOfItems;
			HashMap<Integer, Integer> itemi = (HashMap<Integer, Integer>) items.clone();
			while (itemsLeft != 0){
				boolean wasted = true;
				Grupa grupa = new Grupa(stockWidth, stockHeight);
				ArrayList<Integer> tempPredmeti = new ArrayList<Integer>(itemi.keySet());
				while(grupa.getRemainder() > 0){
					if (tempPredmeti.size() == 0)
						break;
					Integer id = tempPredmeti.get(random.nextInt(tempPredmeti.size()));
					Predmet odabraniPredmet = new Predmet(id, predmeti.get(id).getSirina(), predmeti.get(id).getVisina());
					if (random.nextBoolean())
						odabraniPredmet.rotiraj();
					boolean fits = grupa.addItem(odabraniPredmet);
					if (fits){
						if (itemi.get(id) == 1){
							itemi.remove(id);
							tempPredmeti.remove(id);
						} else
							itemi.put(id, itemi.get(id)-1);
						itemsLeft--;
						if (itemsLeft == 0)
							break;
						continue;
					}
					odabraniPredmet.rotiraj();
					fits = grupa.addItem(odabraniPredmet);
					if (fits){
						if (itemi.get(id) == 1){
							itemi.remove(id);
							tempPredmeti.remove(id);
						} else
							itemi.put(id, itemi.get(id)-1);
						itemsLeft--;
						if (itemsLeft == 0)
							break;
						continue;
					}
					else
						tempPredmeti.remove(id);
				}
				if (grupa.getRemainder() <= maxOstatak)
					wasted = false;
				jedinka.addGroup(grupa, wasted);
			}
			jedinka.calculateCost();
			rjesenja.add(jedinka);
		}
		return rjesenja;
	}
	
	/**
	 * Stvara novu generaciju rješenja. Uzima u obzir elitizam te vjerojatnost nastanka nove jedinke križanjem ili mutacijom
	 */
	public void novaGeneracija(){
		long brojStarih = Math.round(parametri.getVelicinaPopulacije()*parametri.getElitizam());
		if ((parametri.getVelicinaPopulacije() - brojStarih) % 2 == 1)
			brojStarih++;
		Kromosom[] sortiranaPopulacija = populacija.getSortedPopulation();
		ArrayList<Kromosom> novaGeneracija = new ArrayList<Kromosom>();
		for(int i = 0 ; i < brojStarih ; i++ ){
			novaGeneracija.add(sortiranaPopulacija[i]);
		}
		
		for (int i = 0 ; i < (parametri.getVelicinaPopulacije() - brojStarih)/2 ; i++ ){
			Kromosom prviRoditelj = populacija.selektiraj();
			Kromosom drugiRoditelj = populacija.selektiraj();
			while (prviRoditelj.equals(drugiRoditelj)){
				drugiRoditelj = populacija.selektiraj();
			}
			if (random.nextDouble() <= parametri.getVjerojatnostKrizanja()){
				novaGeneracija.add(krizanje(prviRoditelj, drugiRoditelj));
				novaGeneracija.add(krizanje(drugiRoditelj, prviRoditelj));
			} else {
				novaGeneracija.add(mutacija(prviRoditelj));
				novaGeneracija.add(mutacija(drugiRoditelj));
			}
		}
		populacija.setPopulacija(novaGeneracija);
	}
	
	/**
	 * Genetski operator križanja
	 * 
	 * @param prviRoditelj jedinka (prva) od koje će se križanjem stvoriti nova jedinka
	 * @param drugiRoditelj jedinka (druga) od koje će se križanjem stvoriti nova jedinka
	 * @return nova jedinka stvorena križanjem
	 */
	private Kromosom krizanje(Kromosom prviRoditelj, Kromosom drugiRoditelj){
		HashMap<Integer, Integer> itemi = (HashMap<Integer, Integer>) items.clone();
		int itemsLeft = numberOfItems;
		Kromosom dijete = new Kromosom();
		
		int segment = (int) (prviRoditelj.getNumberOfGroups()*0.4);
		Grupa[] roditelj = prviRoditelj.getSolution();
		for(int i = 0 ; i < segment ; i++ ){
			boolean wasted = true;
			
			Grupa grupa = roditelj[i];
			for (Integer j : grupa.getPredmeti().keySet()){
				itemsLeft -= grupa.getPredmeti().get(j);
				if (itemi.get(j) == grupa.getPredmeti().get(j))
					itemi.remove(j);
				else
					itemi.put(j, itemi.get(j)-grupa.getPredmeti().get(j));
			}
			if (grupa.getRemainder() <= maxOstatak)
				wasted = false;
			dijete.addGroup(grupa, wasted);
		}
		roditelj = drugiRoditelj.getSolution();
		for(int i = 0 ; i < roditelj.length ; i++ ){
			boolean wasted = true;
			Grupa grupa = roditelj[i];
			HashMap<Integer, Integer> groupItems = new HashMap<Integer, Integer>();
			for (Integer j : grupa.getPredmeti().keySet()){
				if (groupItems.containsKey(j))
					groupItems.put(j, groupItems.get(j)+grupa.getPredmeti().get(j));
				else
					groupItems.put(j, grupa.getPredmeti().get(j));
			}
			boolean fits = true;
			for (Integer j : groupItems.keySet()){
				if (itemi.get(j) == null || groupItems.get(j) > itemi.get(j)){
					fits = false;
					break;
				}
			}
			if (!fits){
				continue;
			}
			for (Integer j : grupa.getPredmeti().keySet()){
				itemsLeft -= grupa.getPredmeti().get(j);
				if (itemi.get(j) == grupa.getPredmeti().get(j))
					itemi.remove(j);
				else
					itemi.put(j, itemi.get(j)-grupa.getPredmeti().get(j));
			}
			if (grupa.getRemainder() <= maxOstatak)
				wasted = false;
			dijete.addGroup(grupa, wasted);
		}
		
		while (itemsLeft != 0){
			boolean wasted = true;
			Grupa grupa = new Grupa(stockWidth, stockHeight);
			ArrayList<Integer> tempPredmeti = new ArrayList<Integer>(itemi.keySet());
			while(grupa.getRemainder() > 0){
				if (tempPredmeti.size() == 0)
					break;
				Integer id = tempPredmeti.get(random.nextInt(tempPredmeti.size()));
				Predmet odabraniPredmet = new Predmet(id, predmeti.get(id).getSirina(), predmeti.get(id).getVisina());
				if (random.nextBoolean())
					odabraniPredmet.rotiraj();
				boolean fits = grupa.addItem(odabraniPredmet);
				if (fits){
					if (itemi.get(id) == 1){
						itemi.remove(id);
						tempPredmeti.remove(id);
					} else
						itemi.put(id, itemi.get(id)-1);
					itemsLeft--;
					if (itemsLeft == 0)
						break;
					continue;
				}
				odabraniPredmet.rotiraj();
				fits = grupa.addItem(odabraniPredmet);
				if (fits){
					if (itemi.get(id) == 1){
						itemi.remove(id);
						tempPredmeti.remove(id);
					} else
						itemi.put(id, itemi.get(id)-1);
					itemsLeft--;
					if (itemsLeft == 0)
						break;
					continue;
				}
				else
					tempPredmeti.remove(id);
			}
			if (grupa.getRemainder() <= maxOstatak)
				wasted = false;
			dijete.addGroup(grupa, wasted);
		}
		
		dijete.calculateCost();
		return dijete;
	}
	
	/**
	 * Genetski operator mutacije
	 * 
	 * @param roditelj jedinka od koje se mutacijom stvara nova jedinka
	 * @return nova jedinka stvorena mutacijom
	 */
	private Kromosom mutacija(Kromosom roditelj){
		HashMap<Integer, Integer> itemi = (HashMap<Integer, Integer>) items.clone();
		int itemsLeft = numberOfItems;
		Kromosom dijete = new Kromosom();
		for(Grupa i : roditelj.getSolution()){
			boolean wasted = true;
			if (random.nextDouble() <= parametri.getVjerojatnostMutacije())
				continue;
			for (Integer j : i.getPredmeti().keySet()){
				itemsLeft -= i.getPredmeti().get(j);
				if (itemi.get(j) == i.getPredmeti().get(j))
					itemi.remove(j);
				else
					itemi.put(j, itemi.get(j)-i.getPredmeti().get(j));
			}
			if (i.getRemainder() <= maxOstatak)
				wasted = false;
			dijete.addGroup(i, wasted);
		}
		
		while (itemsLeft != 0){
			boolean wasted = true;
			Grupa grupa = new Grupa(stockWidth, stockHeight);
			ArrayList<Integer> tempPredmeti = new ArrayList<Integer>(itemi.keySet());
			while(grupa.getRemainder() > 0){
				if (tempPredmeti.size() == 0)
					break;
				Integer id = tempPredmeti.get(random.nextInt(tempPredmeti.size()));
				Predmet odabraniPredmet = new Predmet(id, predmeti.get(id).getSirina(), predmeti.get(id).getVisina());
				if (random.nextBoolean())
					odabraniPredmet.rotiraj();
				boolean fits = grupa.addItem(odabraniPredmet);
				if (fits){
					if (itemi.get(id) == 1){
						itemi.remove(id);
						tempPredmeti.remove(id);
					} else
						itemi.put(id, itemi.get(id)-1);
					itemsLeft--;
					if (itemsLeft == 0)
						break;
					continue;
				}
				odabraniPredmet.rotiraj();
				fits = grupa.addItem(odabraniPredmet);
				if (fits){
					if (itemi.get(id) == 1){
						itemi.remove(id);
						tempPredmeti.remove(id);
					} else
						itemi.put(id, itemi.get(id)-1);
					itemsLeft--;
					if (itemsLeft == 0)
						break;
					continue;
				}
				else
					tempPredmeti.remove(id);
			}
			if (grupa.getRemainder() <= maxOstatak)
				wasted = false;
			dijete.addGroup(grupa, wasted);
		}
		
		dijete.calculateCost();
		return dijete;
	}
	
	/**
	 * Ispisuje tekstualno rješenje problema
	 * 
	 * @param izlaz tekstualna datoteka u koju se ispisuje rješenje
	 */
	public void ispis(File izlaz){
		try {
			izlaz.setWritable(true, false);
			FileWriter ispis = new FileWriter(izlaz);
			Kromosom rjesenje = populacija.getSortedPopulation()[0];
			for(Grupa i : rjesenje.getSolution()){
				ispis.write(i.toString() + "\n################\n");
			}
			int ukupnoPredmeta = 0;
			for(Integer i : items.keySet()){
				ukupnoPredmeta += (items.get(i)*predmeti.get(i).getSirina()*predmeti.get(i).getVisina());
			}
			int ukupnoMaterijala = rjesenje.getNumberOfGroups()*stockHeight*stockWidth;
			ispis.write("Iskoristenost: " + ((double) ukupnoPredmeta)/ukupnoMaterijala + "\n");
			ispis.flush();
	    	ispis.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Ispisuje sliku rješenja problema
	 * 
	 * @param slika .pdf datoteka u koju se ispisuje slika rješenja
	 */
	public void ispisSlike(File slika){
		Kromosom rjesenje = populacija.getSortedPopulation()[0];
		Document document = new Document();
		try {
			PdfWriter.getInstance(document, 
				new FileOutputStream(slika));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (DocumentException e) {
			e.printStackTrace();
		}
		document.open();
		
		for(Grupa i : rjesenje.getSolution()){
			try {
				com.lowagie.text.Image image = com.lowagie.text.Image.getInstance(i.toPicture(), null);
				image.scaleToFit(document.getPageSize().getWidth() - document.leftMargin() - document.rightMargin(), document.getPageSize().getHeight() - document.topMargin() - document.bottomMargin());
				document.add(image);
				document.newPage();
			} catch (BadElementException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			} catch (DocumentException e) {
				e.printStackTrace();
			}
		}
		document.close();
		
	}
	
	/**
	 * Vraća parametre algoritma
	 * 
	 * @return parametri algoritma
	 */
	public ParametriGA getParametri() {
		return parametri;
	}

	/**
	 * Vraća populaciju rješenja
	 * 
	 * @return populacija rješenja
	 */
	public Populacija getPopulacija() {
		return populacija;
	}

	/**
	 * Funkcija koja pokreće algoritam
	 * 
	 * @param args argumenti pozivanja algoritma
	 */
	public static void main(String[] args) {
		Date pocetak = new Date();
		File parametri = new File("BLFgroupingGA2D_param.txt");
		File ulaz = new File(args[0]);
		File izlaz = new File(args[1]);
		File slika = new File(args[2]);
		BLFGroupingGA2D algoritam = new BLFGroupingGA2D(ulaz, parametri, pocetak.getTime());
		
		int i;
		for (i = 0; i < algoritam.getParametri().getBrojGeneracija() ; i++ ){
			algoritam.novaGeneracija();
		}
		algoritam.ispis(izlaz);
		algoritam.ispisSlike(slika);
	}

}
