import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.HashMap;


public class Grupa implements Comparable<Grupa> {
	
	/** Police na kojima su složeni predmeti */
	private ArrayList<Polica> police;
	/** Predmeti koji se režu od odabranog materijala i njihov broj */
	private HashMap<Integer, Integer> predmeti;
	/** Visina materijala */
	private int stockHeight;
	/** Širina materijala */
	private int stockWidth;
	/** Ostatak nakon rezanja grupe predmeta iz materijala */
	private int remainder;
	/** Ostatak visine materijala za nove police */
	private int ostatakVisine;
	
	/**
	 * Konstruktor objekta
	 * 
	 * @param stockWidth širina materijala
	 * @param stockHeight visina materijala
	 */
	public Grupa(int stockWidth, int stockHeight) {
		super();
		this.stockWidth = stockWidth;
		this.stockHeight = stockHeight;
		police = new ArrayList<Polica>();
		predmeti = new HashMap<Integer, Integer>();
		remainder = stockHeight*stockWidth;
		ostatakVisine = stockHeight;
	}
	
	/**
	 * Pokušava dodati predmet u grupu
	 * 
	 * @param predmet predmet koji se pokušava dodati u grupu
	 * @return true ako je dodavanje uspjelo, inače false
	 */
	public boolean addItem(Predmet predmet){
		for(Polica i : police){
			if (predmet.getVisina() <= i.getVisina() && predmet.getSirina() + i.getZauzetaSirina() <= stockWidth){
				i.add(predmet);
				remainder -= (predmet.getSirina()*predmet.getVisina());
				if (predmeti.containsKey(predmet.getID()))
					predmeti.put(predmet.getID(), predmeti.get(predmet.getID())+1);
				else
					predmeti.put(predmet.getID(), 1);
				return true;
			}
		}
		if (predmet.getVisina() < ostatakVisine){
			ostatakVisine -= predmet.getVisina();
			Polica bla = new Polica(predmet.getVisina());
			bla.add(predmet);
			police.add(bla);
			remainder -= (predmet.getSirina()*predmet.getVisina());
			if (predmeti.containsKey(predmet.getID()))
				predmeti.put(predmet.getID(), predmeti.get(predmet.getID())+1);
			else
				predmeti.put(predmet.getID(), 1);
			return true;
		}
		return false;
	}
	
	/**
	 * Vraća police od kojih se grupa sastoji
	 * 
	 * @return police od kojih se grupa sastoji
	 */
	public ArrayList<Polica> getPolice() {
		return police;
	}

	/**
	 * Vraća duljinu materijala koja ostaje nakon rezanja grupe predmeta iz materijala
	 * 
	 * @return ostatak nakon rezanja
	 */
	public int getRemainder() {
		return remainder;
	}

	/**
	 * Vraća grupu predmeta koji se režu od odabranog materijala
	 * 
	 * @return grupa predmeta
	 */
	public HashMap<Integer, Integer> getPredmeti() {
		return predmeti;
	}

	/* (non-Javadoc)
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo(Grupa grupa) {
		if (this.remainder < grupa.remainder)
			return -1;
		else if (this.remainder > grupa.remainder)
			return 1;
		else
			return 0;
		
	}
	
	@Override
	public String toString(){
		String izlaz = new String();
		int visina = 0;
		for(Polica i : police){
			int sirina = 0;
			for(Predmet j : i.getPredmeti()){
				izlaz += "(" + sirina + "," + visina + ") - (" + (sirina+j.getSirina()) + "," + (visina+j.getVisina()) + ")\n";
				sirina += j.getSirina();
			}
			visina += i.getVisina();
		}
		return izlaz;
	}
	
	/**
	 * Stvara sliku rješenja
	 * 
	 * @return slika rješenja
	 */
	public BufferedImage toPicture(){
		int sirinaMjerila = 35;
		BufferedImage image = new BufferedImage(stockWidth + sirinaMjerila + 15, stockHeight + sirinaMjerila + 10, BufferedImage.TYPE_BYTE_INDEXED);
    	Graphics2D graphics = image.createGraphics();
    	graphics.setBackground(Color.white);
    	graphics.clearRect(0, 0, stockWidth + sirinaMjerila + 15, stockHeight + sirinaMjerila + 10);
    	
    	graphics.setColor(Color.black);
    	int korakMjerila = 100;
    	int brojac = 0;
    	for(double x = sirinaMjerila ; (x - 0.1) <= (stockWidth + sirinaMjerila) ; x += korakMjerila){
    		if (brojac % 5 == 0) {
    			graphics.draw(new Line2D.Double(x, sirinaMjerila, x, sirinaMjerila - 10));
    			
    		} else {
    			graphics.draw(new Line2D.Double(x, sirinaMjerila, x, sirinaMjerila - 5));
    		}
    		graphics.drawString(String.valueOf(brojac * korakMjerila), (float) (x - 15), sirinaMjerila - 15);
    		brojac++;
    	}
    	brojac = 0;
    	for(double y = sirinaMjerila ; (y - 0.1) <= (stockHeight + sirinaMjerila) ; y += korakMjerila){
    		if (brojac % 5 == 0) {
    			graphics.draw(new Line2D.Double(sirinaMjerila, y, sirinaMjerila - 10, y));
    			
    		} else {
    			graphics.draw(new Line2D.Double(sirinaMjerila, y, sirinaMjerila - 5, y));
    		}
    		graphics.drawString(String.valueOf(brojac * korakMjerila), (float) (sirinaMjerila - 35), (float) (y + 5));
    		brojac++;
    	}
    	
    	Rectangle2D.Double rect;
    	int trenutnaVisina = 0;
    	for(Polica i : police){
    		int trenutnaSirina = 0;
    		for(Predmet j : i.getPredmeti()){
    			rect = new Rectangle2D.Double(trenutnaSirina + sirinaMjerila, stockHeight - trenutnaVisina - j.getVisina() + sirinaMjerila, j.getSirina(), j.getVisina());
        		graphics.setColor(Color.cyan);
    			graphics.fill(rect);
    			graphics.setColor(Color.blue);
    			graphics.draw(rect);
    			trenutnaSirina += j.getSirina();
    		}
    		trenutnaVisina += i.getVisina();
    	}
    	graphics.dispose();
		return image;
	}

}
