import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Random;


public class HybridGroupingGA {
	
	/** Debljina rezne ploče, stvara gubitke prilikom rezanja */
	private int debljinaReza;
	/** Maksimalna duljina ostatka za koji se materijal smatra dovoljno dobro izrezanim */
	private int maxOstatak;
	/** Popis predmeta koji je potrebno izrezati */
	private HashMap<Integer, Integer> items;
	/** Popis raspoloživih materijala */
	private HashMap<Integer, Integer> stock;
	/** Parametri genetskog algoritma */
	private ParametriGA parametri;
	/** Trenutna populacija rješenja genetskog algoritma */
	private Populacija populacija;
	/** Ukupni broj predmeta koji je potrebno izrezati */
	private int numberOfItems;
	/** Generator nasumičnih brojeva */
	private Random random;
	/** Maksimalno vrijeme izvođenja algoritma */
	private long maxTime;
	
	/**
	 * Konstruktor objekta
	 * 
	 * @param ulaz datoteka sa ulaznim podacima
	 * @param parametri datoteka sa parametrima algoritma
	 * @param time trenutno vrijeme, dobiveno iz Date objekta
	 */
	public HybridGroupingGA(File ulaz, File parametri, long time){
		super();
		try {
			this.items = new HashMap<Integer, Integer>();
			this.stock = new HashMap<Integer, Integer>();
			this.random = new Random(time);
			unos(ulaz);
			this.parametri = new ParametriGA(parametri);
		} catch (IOException e) {
			e.printStackTrace();
		}
		this.populacija = new Populacija(generiranjePopulacije());
	}
	
	/**
	 * Učitava podatke iz ulazne datoteke
	 * 
	 * @param ulaz ulazna datoteka iz koje se učitavaju ulazni podaci
	 * @throws IOException greška prilikom čitanja iz datoteke
	 */
	private void unos(File ulaz) throws IOException{
		BufferedReader citac = new BufferedReader(new FileReader(ulaz));
		String red = citac.readLine();
		String[] temp = red.split("\t");
		maxTime = Integer.parseInt(temp[0]);
		debljinaReza = Integer.parseInt(temp[1]);
		maxOstatak = Integer.parseInt(temp[2]);
		red = citac.readLine();
		while (!(red.charAt(0) == '-')){
			temp = red.split("\t");
			stock.put(Integer.valueOf(temp[0]), Integer.valueOf(temp[1]));
			red = citac.readLine();
		}
		numberOfItems = 0;
		red = citac.readLine();
		while (red != null) {
			temp = red.split("\t");
			items.put(Integer.valueOf(temp[0]), Integer.valueOf(temp[1]));
			numberOfItems += Integer.valueOf(temp[1]);
			red = citac.readLine();
		}
	}
	
	/**
	 * Funkcija koja postupkom slučajnog odabira generira jedinke, te tako stvara početnu populaciju
	 * 
	 * @return generirana populacija
	 */
	private ArrayList<HKromosom> generiranjePopulacije(){
		ArrayList<HKromosom> rjesenja = new ArrayList<HKromosom>();
		for (int i = 0 ; i < parametri.getVelicinaPopulacije(); i++ ){
			HKromosom jedinka = new HKromosom();
			int itemsLeft = numberOfItems;
			Integer[] duljineMaterijala = stock.keySet().toArray(new Integer[0]);
			HashMap<Integer, Integer> materijal = (HashMap<Integer, Integer>) stock.clone();
			HashMap<Integer, Integer> predmeti = (HashMap<Integer, Integer>) items.clone();
			
			while (itemsLeft != 0){
				boolean newStock = false;
				boolean wasted = true;
				int stockLength = duljineMaterijala[random.nextInt(duljineMaterijala.length)];
				int remainder = stockLength;
				if (materijal.get(stockLength) == -1)
					newStock = true;
				else{
					if (materijal.get(remainder) == 1){
						materijal.remove(remainder);
						duljineMaterijala = materijal.keySet().toArray(new Integer[0]);
					} else
						materijal.put(stockLength, materijal.get(stockLength)-1);
				}
				
				ArrayList<Integer> grupa = new ArrayList<Integer>();
				while(remainder > 0){
					ArrayList<Integer> duljinePredmeta = new ArrayList<Integer>();
					for(Integer j : predmeti.keySet()){
						if (j <= remainder)
							duljinePredmeta.add(j);
					}
					if (duljinePredmeta.size() == 0)
						break;
					int duljina = duljinePredmeta.get(random.nextInt(duljinePredmeta.size()));
					grupa.add(duljina);
					if (predmeti.get(duljina) == 1)
						predmeti.remove(duljina);
					else
						predmeti.put(duljina, predmeti.get(duljina)-1);
					itemsLeft--;
					if (itemsLeft == 0)
						break;
					remainder = remainder - duljina - debljinaReza;
					if (remainder < 0)
						remainder = 0;
				}
				if (grupa.size() == 0){
					if (materijal.containsKey(stockLength)){
						if (materijal.get(stockLength) != -1)
							materijal.put(stockLength, materijal.get(stockLength)+1);
					} else
						materijal.put(stockLength, 1);
					continue;
				}
				if (remainder <= maxOstatak)
					wasted = false;
				HGrupa tmp = new HGrupa(grupa, stockLength, remainder);
				jedinka.addGroup(tmp, newStock, wasted);
			}
			jedinka.calculateCost();
			rjesenja.add(jedinka);
		}
		return rjesenja;
	}

	/**
	 * Stvara novu generaciju rješenja. Uzima u obzir elitizam te vjerojatnost nastanka nove jedinke križanjem ili mutacijom
	 */
	public void novaGeneracija(){
		long brojStarih = Math.round(parametri.getVelicinaPopulacije()*parametri.getElitizam());
		if ((parametri.getVelicinaPopulacije() - brojStarih) % 2 == 1)
			brojStarih++;
		HKromosom[] sortiranaPopulacija = populacija.getSortedPopulation();
		ArrayList<HKromosom> novaGeneracija = new ArrayList<HKromosom>();
		for(int i = 0 ; i < brojStarih ; i++ ){
			novaGeneracija.add(sortiranaPopulacija[i]);
		}
		
		for (int i = 0 ; i < (parametri.getVelicinaPopulacije() - brojStarih)/2 ; i++ ){
			HKromosom prviRoditelj = populacija.selektiraj();
			HKromosom drugiRoditelj = populacija.selektiraj();
			while (prviRoditelj.equals(drugiRoditelj)){
				drugiRoditelj = populacija.selektiraj();
			}
			if (random.nextDouble() <= parametri.getVjerojatnostKrizanja()){
				novaGeneracija.add(krizanje(prviRoditelj, drugiRoditelj));
				novaGeneracija.add(krizanje(drugiRoditelj, prviRoditelj));
			} else {
				novaGeneracija.add(mutacija(prviRoditelj));
				novaGeneracija.add(mutacija(drugiRoditelj));
			}
		}
		populacija.setPopulacija(novaGeneracija);
	}
	
	/**
	 * Genetski operator križanja sa lokalnom pretragom
	 * 
	 * @param prviRoditelj jedinka (prva) od koje će se križanjem stvoriti nova jedinka
	 * @param drugiRoditelj jedinka (druga) od koje će se križanjem stvoriti nova jedinka
	 * @return nova jedinka stvorena križanjem
	 */
	private HKromosom krizanje(HKromosom prviRoditelj, HKromosom drugiRoditelj){
		HashMap<Integer, Integer> materijal = (HashMap<Integer, Integer>) stock.clone();
		HashMap<Integer, Integer> predmeti = (HashMap<Integer, Integer>) items.clone();
		int itemsLeft = numberOfItems;
		HKromosom dijete = new HKromosom();
		int segment = (int) (prviRoditelj.getNumberOfGroups()*0.4);
		HGrupa[] roditelj = prviRoditelj.getSortedSolution();
		for(int i = 0 ; i < segment ; i++ ){
			boolean newStock = false;
			boolean wasted = true;
			
			HGrupa grupa = roditelj[i];
			if (materijal.get(grupa.getStockLength()) == -1)
				newStock = true;
			else {
				if (materijal.get(grupa.getStockLength()) == 1)
					materijal.remove(grupa.getStockLength());
				else
					materijal.put(grupa.getStockLength(), materijal.get(grupa.getStockLength())-1);
			}
			for (Integer j : grupa.getGroup()){
				itemsLeft--;
				if (predmeti.get(j) == 1)
					predmeti.remove(j);
				else
					predmeti.put(j, predmeti.get(j)-1);
			}
			if (grupa.getRemainder() <= maxOstatak)
				wasted = false;
			dijete.addGroup(new HGrupa((ArrayList<Integer>) grupa.getGroup().clone(), grupa.getStockLength(), grupa.getRemainder()), newStock, wasted);
		}
		roditelj = drugiRoditelj.getSortedSolution();
		for(int i = 0 ; i < roditelj.length ; i++ ){
			boolean newStock = false;
			boolean wasted = true;
			HGrupa grupa = roditelj[i];
			HashMap<Integer, Integer> groupItems = new HashMap<Integer, Integer>();
			for (Integer j : grupa.getGroup()){
				if (groupItems.containsKey(j))
					groupItems.put(j, groupItems.get(j)+1);
				else
					groupItems.put(j, 1);
			}
			boolean fits = true;
			for (Integer j : groupItems.keySet()){
				if (predmeti.get(j) == null || groupItems.get(j) > predmeti.get(j)){
					fits = false;
					break;
				}
			}
			if (!fits){
				continue;
			}
			if (!materijal.containsKey(grupa.getStockLength())){
				continue;
			}
			if (materijal.get(grupa.getStockLength()) == -1)
				newStock = true;
			else {
				if (materijal.get(grupa.getStockLength()) == 1)
					materijal.remove(grupa.getStockLength());
				else
					materijal.put(grupa.getStockLength(), materijal.get(grupa.getStockLength())-1);
			}
			for (Integer j : grupa.getGroup()){
				itemsLeft--;
				if (predmeti.get(j) == 1)
					predmeti.remove(j);
				else
					predmeti.put(j, predmeti.get(j)-1);
			}
			if (grupa.getRemainder() <= maxOstatak)
				wasted = false;
			dijete.addGroup(new HGrupa((ArrayList<Integer>) grupa.getGroup().clone(), grupa.getStockLength(), grupa.getRemainder()), newStock, wasted);
		}
		
		//local optimisation using dominance criterion (two for one)
		for(HGrupa i : dijete.getSolution()){
			boolean mijenjaj = false;
			Integer prviItem = 0;
			Integer drugiItem = 0;
			int duljinaPredmeta = 0;
			for(Integer j : i.getGroup()){
				if (mijenjaj || i.getRemainder() < maxOstatak)
					break;
				for(Integer k : i.getGroup()){
					if (mijenjaj)
						break;
					if (j.equals(k))
						continue;
					int zbroj = j + k + debljinaReza;//dodajem debljinu reza jer ce biti jedan rez manje
					for(Integer l : predmeti.keySet()){
						if (l > zbroj && l - zbroj < (i.getRemainder() + debljinaReza)){
							prviItem = j;
							drugiItem = k;
							duljinaPredmeta = l;
							mijenjaj = true;
						}
					}
				}
			}
			if (mijenjaj){
				boolean wasted = false;
				if (i.getRemainder() > maxOstatak)
					wasted = true;
				i.removeFromGroup(prviItem, debljinaReza);
				i.removeFromGroup(drugiItem, debljinaReza);
				i.addToGroup(duljinaPredmeta, debljinaReza);
				if (wasted && i.getRemainder() < maxOstatak)
					dijete.smanjiWasted();
				if (predmeti.get(duljinaPredmeta) == 1)
					predmeti.remove(duljinaPredmeta);
				else
					predmeti.put(duljinaPredmeta, predmeti.get(duljinaPredmeta)-1);
				
				if (predmeti.containsKey(prviItem))
					predmeti.put(prviItem, predmeti.get(prviItem)+1);
				else
					predmeti.put(prviItem, 1);
				
				if (predmeti.containsKey(drugiItem))
					predmeti.put(drugiItem, predmeti.get(drugiItem)+1);
				else
					predmeti.put(drugiItem, 1);
				itemsLeft++;
			}
		}
		
		Integer[] duljineMaterijala = materijal.keySet().toArray(new Integer[0]);
		while (itemsLeft != 0){
			boolean newStock = false;
			boolean wasted = true;
			int stockLength = duljineMaterijala[random.nextInt(duljineMaterijala.length)];
			int remainder = stockLength;
			if (materijal.get(stockLength) == -1)
				newStock = true;
			else{
				if (materijal.get(remainder) == 1){
					materijal.remove(remainder);
					duljineMaterijala = materijal.keySet().toArray(new Integer[0]);
				} else
					materijal.put(stockLength, materijal.get(stockLength)-1);
			}
			
			ArrayList<Integer> grupa = new ArrayList<Integer>();
			while(remainder > 0){
				ArrayList<Integer> duljinePredmeta = new ArrayList<Integer>();
				for(Integer j : predmeti.keySet()){
					if (j <= remainder)
						duljinePredmeta.add(j);
				}
				if (duljinePredmeta.size() == 0)
					break;
				int duljina = duljinePredmeta.get(random.nextInt(duljinePredmeta.size()));
				grupa.add(duljina);
				if (predmeti.get(duljina) == 1){
					predmeti.remove(duljina);
				} else
					predmeti.put(duljina, predmeti.get(duljina)-1);
				remainder = remainder - duljina - debljinaReza;
				if (remainder < 0)
					remainder = 0;
				itemsLeft--;
				if (itemsLeft == 0)
					break;
			}
			if (grupa.size() == 0){
				if (materijal.containsKey(stockLength)){
					if (materijal.get(stockLength) != -1)
						materijal.put(stockLength, materijal.get(stockLength)+1);
				} else
					materijal.put(stockLength, 1);
				continue;
			}
			if (remainder <= maxOstatak)
				wasted = false;
			dijete.addGroup(new HGrupa(grupa, stockLength, remainder), newStock, wasted);
		}
		
		dijete.calculateCost();
		return dijete;
	}
	
	/**
	 * Genetski operator mutacije sa lokalnom pretragom
	 * 
	 * @param roditelj jedinka od koje se mutacijom stvara nova jedinka
	 * @return nova jedinka stvorena mutacijom
	 */
	private HKromosom mutacija(HKromosom roditelj){
		HashMap<Integer, Integer> materijal = (HashMap<Integer, Integer>) stock.clone();
		HashMap<Integer, Integer> predmeti = (HashMap<Integer, Integer>) items.clone();
		int itemsLeft = numberOfItems;
		HKromosom dijete = new HKromosom();
		for(HGrupa i : roditelj.getSolution()){
			boolean newStock = false;
			boolean wasted = true;
			if (random.nextDouble() <= parametri.getVjerojatnostMutacije())
				continue;
			if (materijal.get(i.getStockLength()) == -1)
				newStock = true;
			else {
				if (materijal.get(i.getStockLength()) == 1)
					materijal.remove(i.getStockLength());
				else
					materijal.put(i.getStockLength(), materijal.get(i.getStockLength())-1);
			}
			for (Integer j : i.getGroup()){
				itemsLeft--;
				if (predmeti.get(j) == 1)
					predmeti.remove(j);
				else
					predmeti.put(j, predmeti.get(j)-1);
			}
			if (i.getRemainder() <= maxOstatak)
				wasted = false;
			dijete.addGroup(new HGrupa((ArrayList<Integer>) i.getGroup().clone(), i.getStockLength(), i.getRemainder()), newStock, wasted);
		}
		
		//local optimisation using dominance criterion (two for one)
		for(HGrupa i : dijete.getSolution()){
			boolean mijenjaj = false;
			Integer prviItem = 0;
			Integer drugiItem = 0;
			int duljinaPredmeta = 0;
			for(Integer j : i.getGroup()){
				if (mijenjaj || i.getRemainder() < maxOstatak)
					break;
				for(Integer k : i.getGroup()){
					if (mijenjaj)
						break;
					if (j.equals(k))
						continue;
					int zbroj = j + k + debljinaReza;//dodajem debljinu reza jer ce biti jedan rez manje
					for(Integer l : predmeti.keySet()){
						if (l > zbroj && l - zbroj < (i.getRemainder() + debljinaReza)){
							prviItem = j;
							drugiItem = k;
							duljinaPredmeta = l;
							mijenjaj = true;
						}
					}
				}
			}
			if (mijenjaj){
				boolean wasted = false;
				if (i.getRemainder() > maxOstatak)
					wasted = true;
				i.removeFromGroup(prviItem, debljinaReza);
				i.removeFromGroup(drugiItem, debljinaReza);
				i.addToGroup(duljinaPredmeta, debljinaReza);
				if (wasted && i.getRemainder() < maxOstatak)
					dijete.smanjiWasted();
				if (predmeti.get(duljinaPredmeta) == 1)
					predmeti.remove(duljinaPredmeta);
				else
					predmeti.put(duljinaPredmeta, predmeti.get(duljinaPredmeta)-1);
				
				if (predmeti.containsKey(prviItem))
					predmeti.put(prviItem, predmeti.get(prviItem)+1);
				else
					predmeti.put(prviItem, 1);
				
				if (predmeti.containsKey(drugiItem))
					predmeti.put(drugiItem, predmeti.get(drugiItem)+1);
				else
					predmeti.put(drugiItem, 1);
				itemsLeft++;
			}
		}
		
		Integer[] duljineMaterijala = materijal.keySet().toArray(new Integer[0]);
		while (itemsLeft != 0){
			boolean newStock = false;
			boolean wasted = true;
			int stockLength = duljineMaterijala[random.nextInt(duljineMaterijala.length)];
			int remainder = stockLength;
			if (materijal.get(stockLength) == -1)
				newStock = true;
			else{
				if (materijal.get(remainder) == 1){
					materijal.remove(remainder);
					duljineMaterijala = materijal.keySet().toArray(new Integer[0]);
				} else
					materijal.put(stockLength, materijal.get(stockLength)-1);
			}
			
			ArrayList<Integer> grupa = new ArrayList<Integer>();
			while(remainder > 0){
				ArrayList<Integer> duljinePredmeta = new ArrayList<Integer>();
				for(Integer j : predmeti.keySet()){
					if (j <= remainder)
						duljinePredmeta.add(j);
				}
				if (duljinePredmeta.size() == 0)
					break;
				int duljina = duljinePredmeta.get(random.nextInt(duljinePredmeta.size()));
				grupa.add(duljina);
				if (predmeti.get(duljina) == 1)
					predmeti.remove(duljina);
				else
					predmeti.put(duljina, predmeti.get(duljina)-1);
				remainder = remainder - duljina - debljinaReza;
				if (remainder < 0)
					remainder = 0;
				itemsLeft--;
				if (itemsLeft == 0)
					break;
				
			}
			if (grupa.size() == 0){
				if (materijal.containsKey(stockLength)){
					if (materijal.get(stockLength) != -1)
						materijal.put(stockLength, materijal.get(stockLength)+1);
				} else
					materijal.put(stockLength, 1);
				continue;
			}
			if (remainder <= maxOstatak)
				wasted = false;
			
			dijete.addGroup(new HGrupa(grupa, stockLength, remainder), newStock, wasted);
		}
		
		dijete.calculateCost();
		return dijete;
	}
	
	/**
	 * Ispisuje rješenje problema
	 * 
	 * @param izlaz datoteka u koju se ispisuje rješenje
	 */
	public void ispis(File izlaz){
		try {
			izlaz.setWritable(true, false);
			FileWriter ispis = new FileWriter(izlaz, true);
			HKromosom rjesenje = populacija.getSortedPopulation()[0];
			double ukupnoMaterijala = 0;
			double ukupnoPredmeta = 0;
			for(Integer i : items.keySet()){
				ukupnoPredmeta += (i*items.get(i));
			}
			for (HGrupa i : rjesenje.getSolution()){
				ukupnoMaterijala += i.getStockLength();
				ispis.write(i.toString() + "\n");
			}
			ispis.write("ukupno iskoristenost: " + String.valueOf(ukupnoPredmeta/ukupnoMaterijala) + "\n\n");
			ispis.flush();
	    	ispis.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Vraća parametre algoritma
	 * 
	 * @return parametri algoritma
	 */
	public ParametriGA getParametri() {
		return parametri;
	}

	/**
	 * Vraća populaciju rješenja
	 * 
	 * @return populacija rješenja
	 */
	public Populacija getPopulacija() {
		return populacija;
	}

	/**
	 * Vraća maksimalno vrijeme izvođenja
	 * 
	 * @return maksimalno vrijeme izvođenja
	 */
	public long getMaxTime() {
		return maxTime;
	}

	/**
	 * Funkcija koja pokreće algoritam
	 * 
	 * @param args argumenti pozivanja algoritma
	 */
	public static void main(String[] args) {
		Date pocetak = new Date();
		File parametri = new File("HgroupingGA_param.txt");
		File ulaz = new File(args[0]);
		File izlaz = new File(args[1]);
		HybridGroupingGA algoritam = new HybridGroupingGA(ulaz, parametri, pocetak.getTime());
		
		for (int i = 0; i < algoritam.getParametri().getBrojGeneracija() ; i++ ){
			Date sad = new Date();
			if (sad.getTime() - pocetak.getTime() > algoritam.getMaxTime()*1000)
				break;
			algoritam.novaGeneracija();
		}
		algoritam.ispis(izlaz);
	}

}
